{-# OPTIONS_GHC -Wall #-}
{-# LANGUAGE FlexibleInstances                #-}

module Main where

-- Libs
import System.Console.ParseArgs hiding (args) 
import System.Environment

-- Music stuff
import HarmTrace.HarmTrace
import HarmTrace.IO.Main
import HarmTrace.IO.Errors

--------------------------------------------------------------------------------
-- FHarm imports
--------------------------------------------------------------------------------
import           HarmTrace.Base.MusicRep
import           HarmTrace.Harmony.ChordRep
import           HarmTrace.Harmony.HaskoreBridge
import           Data.List
import           Data.Maybe   
import           System.Random     
import           HarmTrace.HAnTree.ToHAnTree 
import           Data.Ord (comparing)         
import           HarmTrace.HAnTree.Tree    
import           HarmTrace.HAnTree.HAn
import           Data.Ratio

--------------------------------------------------------------------------------
-- FHarm functions 
--------------------------------------------------------------------------------

type Harmony    = [ChordLabel]
type Octave     = Int   

modKeys :: [Key]
modKeys = [Key (Note (Just m) n ) a | m <- [Sh .. FF], n <-[C .. B], a <- [MajMode .. MinMode]]
normKeys :: [Key]
normKeys = [Key (Note (Nothing) n ) a | n <-[C .. B], a <- [MajMode .. MinMode]]
allKeys :: [Key]
allKeys = sort(modKeys ++ normKeys)

main :: IO ()
main = do args <- getArgs
          case length(args) of
            4           -> harmonise (read (args!!0)::Int) (read (args!!1)::Int) (inputKey (args!!2)) (args!!3)
            _           -> putStrLn(err5)

-- for ghci use
iharmonise :: Int -> Int -> [Char] -> FilePath -> IO ()
iharmonise amountOfHarms y k path = harmonise amountOfHarms y (inputKey k) path

-- Usage: "harmonize [number of harmonizations as integer number] [key of the piece in HarmTrace notation] [path of input midi file as string]"
-- fHarm needs the key in HarmTrace notation because of a bug in Haskore
harmonise :: Int -> Int -> Key -> FilePath -> IO ()
harmonise amountOfHarms y k path = do maybeconvert <- midi2HarmTrace path
                                      if (maybeconvert == Nothing) then do
                                         --error ("harmonize: Error while opening file: " ++ path)
                                         _ <- error ((++) "harmonize: Error while opening file: " path)
                                         return()
                                         else do let song = fromJust maybeconvert
                                                 let (Song key t s m) = song    
                                                 -- convert melody to relative ints
                                                 let relativeMelody = barNotesToInts k m   
                                                 putStrLn ("--> relativeMelody: " ++ show relativeMelody)                                    
                                                 -- make the chordlist with every bar
                                                 let withChords = findCandidateChords k relativeMelody
                                                 -- only chords every yth beat
                                                 let skipped = filterSkipNotes t y withChords 
                                                 putStrLn ("--> skipped: " ++ (show skipped))                                          

                                                 -- give chords their propabilities
                                                 let propped = propChords k skipped
                                                
                                                 putStrLn ("--> Using file: " ++ path)                                    
                                                 putStrLn ("--> Key: " ++ (keyToString k)) 
                                                 putStrLn ("--> Time signature: " ++ (show t))
                                                 putStrLn ("--> Melody:") 
                                                 pretttyPrintMel propped         
                                                
                                                 -- generate n random sequences  
                                                 putStrLn $ "--> Creating " ++ (show amountOfHarms) ++ " harmonizations... "
                                                 sequences <- generateNRandomSequences amountOfHarms propped
                                                 -- putStrLn (show sequences)    


                                                 let noImps =  map filterImpChords sequences
                                                 let asRoots = (map.map) (chordScaleDegreeToRoot k) noImps         
                                                 -- putStrLn (show asRoots)
                                                 let grammar = Jazz
                                                 let options = defaultOpts                                                                       
                                                 
                                                 -- as strings:
                                                 let stringSeqs = map ((++) (keyToString k ++ ";")) $ init $ map sequenceAsString asRoots
                                                
                                                 -- parse the strings all into a  list [ParseResult Piece]
                                                 putStrLn $ "--> Parsing " ++ (show amountOfHarms) ++ " harmonizations... "
                                                 parsedHarmonies <- mapM (parseTreeSilent grammar options) stringSeqs
                                                
                                                 -- get the best of the parsed harmonies  
                                                 bestErrorTrees <- getNBestParseResult 5 parsedHarmonies 
                                                
                                                 putStrLn "--> 5 Best harmonizations: "
                                                 -- show n best harmonisations
                                                 showAllParseResults 0 bestErrorTrees
                                                
                                                 let b@(ParseResult k1 chrd list tree ambtrees tokerr pieceerr pproc) = head bestErrorTrees
                                                 --printParseresult b         
                                                 putStrLn ("--> Chords: " ++ (show chrd))
                                                
                                                 -- make a midi file from the melody and harmonization
                                                 putStrLn ("--> Writing midi file: " ++ (filenameAddHarm path))
                                                 toMidi t y (filenameAddHarm path) m chrd
                                                                                               
                                                 return()                                              

-- adds "_harmonized" and the current time to the filename
filenameAddHarm :: FilePath -> FilePath
filenameAddHarm path | ((drop (length(path)-4) path) == ".mid")  = ((take (length(path)-4) path) ++ "_harmonized.mid")
                     | ((drop (length(path)-5) path) == ".midi") = ((take (length(path)-5) path) ++ "_harmonized.midi")
                     | otherwise                                  = (path ++ "_harmonized.midi")
                     
printParseresult :: ParseResult g -> IO ()
printParseresult p@(ParseResult k chrd list tree ambtrees tokerr pieceerr pproc) = 
        do putStrLn ("Key : " ++ (show k))
           putStrLn ("Chords : " ++ (show chrd))
           putStrLn ("countErrors : " ++ (show $ countErrors pieceerr))
           putStrLn ("Tree : " ++ (show chrd))           
           putStrLn ("Func Nodes : " ++ (show $ getNoFuncNodes p))           
           putStrLn ("Ambtrees : " ++ (show ambtrees))
           putStrLn ("Token err : " ++ (show $ length tokerr))
           putStrLn ("Parser err : " ++ (show $ length pieceerr))  
           putStrLn ("postProcessing : " ++ (show postProcessing))
           return()
           
getByLength :: [ParseResult g] -> [ParseResult g]
getByLength [] = []
getByLength (p@(ParseResult k chrd list tree ambtrees tokerr pieceerr pproc):xs) | ((length chrd) == getNrFuncNodes tree) = p:getByLength xs
                                                                                 | otherwise                              = getByLength xs

-- returns the n best parse results by 
--   first sorting on the least amount of general errors
getNBestParseResult :: (GTree g) => Int -> [ParseResult g] -> IO [ParseResult g]
getNBestParseResult n a = do putStrLn ("--> Sorting harmonizations...")
                             let bestErrorTrees = take n $ sortBy sortParseResult3 a
                             let treesByNoNodes = sortBy (comparing getNoFuncNodes) bestErrorTrees
                             return(treesByNoNodes)          

getNoFuncNodes :: (ParseResult a) -> Int
getNoFuncNodes (ParseResult _ _ _ t _ _ _ _) = getNoFuncNode t where
        getNoFuncNode :: Tree HAn -> Int
        getNoFuncNode (Node (HAnFunc P) nodes _) = length nodes
        getNoFuncNode _ = error "HarmTrace.hs: not a correctly formed HAn Tree"                              

getBestParseResult :: [ParseResult g] -> ParseResult g
getBestParseResult a = head $ sortBy sortParseResult a                                                        

-- sort by deletions
sortParseResult :: ParseResult t -> ParseResult t1 -> Ordering
sortParseResult (ParseResult k chrd list tree ambtrees tokerr pieceerr pproc) (ParseResult k2 chrd2 list2 tree2 ambtrees2 tokerr2 pieceerr2 pproc2) 
        | d < d2 = GT   
        | d >= d2 = LT  where 
                (ErrorNrs i d de r) = countErrors pieceerr
                (ErrorNrs i2 d2 de2 r2) = countErrors pieceerr2

-- sort by insertions
sortParseResult2 :: ParseResult t -> ParseResult t1 -> Ordering
sortParseResult2 (ParseResult k chrd list tree ambtrees tokerr pieceerr pproc) (ParseResult k2 chrd2 list2 tree2 ambtrees2 tokerr2 pieceerr2 pproc2) 
        | (length chrd) >  (length chrd2) = GT   
        | (length chrd) <= (length chrd2) = LT  where 
                (ErrorNrs i d de r) = countErrors pieceerr
                (ErrorNrs i2 d2 de2 r2) = countErrors pieceerr2   

-- sort by simple error simpleErrorMeasure
sortParseResult3 :: ParseResult t -> ParseResult t1 -> Ordering
sortParseResult3 (ParseResult k chrd list tree ambtrees tokerr pieceerr pproc) (ParseResult k2 chrd2 list2 tree2 ambtrees2 tokerr2 pieceerr2 pproc2) 
        | (simpleErrorMeasure e1) >  (simpleErrorMeasure e2) = GT   
        | (simpleErrorMeasure e1) <= (simpleErrorMeasure e2) = LT  where 
                e1 = countErrors pieceerr
                e2 = countErrors pieceerr2        
                
showAllParseResults :: (GTree g) => Int -> [ParseResult g] -> IO()
showAllParseResults _ []     = return()
showAllParseResults n (x:xs) = do putStr $ show n
                                  mapM_ (print . gTreeHead) (parsedPiece x) 
                                  printParseresult x
                                  showAllParseResults (n+1) xs
                                  return()

sequenceAsString :: [Chord Root] -> String          
sequenceAsString []                                 = ""
sequenceAsString [(Chord (Note (Nothing) r) sh l i d)]    = ((show r) ++ ":" ++ (shortHandAsString sh))
sequenceAsString ((Chord (Note (Nothing) r) sh l i d):xs) = ((show r) ++ ":" ++ (shortHandAsString sh)) ++ ";" ++ sequenceAsString xs
sequenceAsString [(Chord (Note (m) r) sh l i d)]    = ((show r) ++ show(fromJust m) ++ ":" ++ (shortHandAsString sh))
sequenceAsString ((Chord (Note (m) r) sh l i d):xs) = ((show r) ++ show(fromJust m) ++ ":" ++ (shortHandAsString sh)) ++ ";" ++ sequenceAsString xs


shortHandAsString :: Shorthand -> String
shortHandAsString  Maj     = "maj"
shortHandAsString  Min     = "min"
shortHandAsString  Dim     = "dim"
shortHandAsString  Aug     = "aug"
shortHandAsString  Maj7    = "maj7"
shortHandAsString  Min7    = "min7"
shortHandAsString  Sev     = "7"
shortHandAsString  Dim7    = "dim7"
shortHandAsString  HDim7   = "hdim"
shortHandAsString  MinMaj7 = "minmaj7"
shortHandAsString  Maj6    = "maj6"
shortHandAsString  Min6    = "min6"
shortHandAsString  Nin     = "9"
shortHandAsString  Maj9    = "maj9"
shortHandAsString  Min9    = "min9"
shortHandAsString  Sus4    = "sus4" 
shortHandAsString  Sus2    = "sus2" 
shortHandAsString  None    = ""

filterImpChords :: [Chord ScaleDegree] -> [Chord ScaleDegree]
filterImpChords []     = []
filterImpChords (x:xs) | (x == (Chord (Note (Nothing) Imp) None [] (-9) (-9))) = filterImpChords xs
                       | otherwise                                             = x:filterImpChords xs
                   
generateNRandomSequences :: Int -> [(Bar,[(Int,Octave,Duration,PropChordList)])] -> IO [[Chord ScaleDegree]]
generateNRandomSequences 0 l  = return([[]])
generateNRandomSequences _ [] = return([[]])
generateNRandomSequences n l  = do chordSequence <- generateRandomSequence l
                                   chordList <- generateNRandomSequences (n-1) l
                                   return(chordSequence:chordList)

generateRandomSequence :: [(Bar,[(Int,Octave,Duration,PropChordList)])] -> IO [Chord ScaleDegree]
generateRandomSequence [] = return([])
generateRandomSequence ((b,m):ms) = do chords <- pickChord m 
                                       nextchord <- generateRandomSequence ms
                                       return(chords ++ nextchord)

pickChord :: [(Int,Octave,Duration,PropChordList)] -> IO [Chord ScaleDegree]
pickChord []             = return([])
pickChord ((i,o,d,l):ms) = do a <- (sample (sortPropChordList (transf l)))
                              b <- pickChord ms
                              return(a:b)

sample :: [(Chord ScaleDegree, Double)] -> IO (Chord ScaleDegree)
sample [] = return(Chord (Note (Nothing) Imp) None [] (-9) (-9))
sample l  = do  p <- getStdRandom (randomR (0.0,1.0)) 
                return (fst(l!!(pick p l)))

pick :: Double -> [(Chord ScaleDegree, Double)] -> Int
pick p  =  fromJust . findIndex (> p) . tail . scanl (\a b -> a + snd b) 0

--pick :: PropChordList -> IO (Chord ScaleDegree)
--pick []                        = return(Chord (Note (Nothing) Imp) None [] (-9) (-9))
--pick [(c1,p1),(c2,p2),(c3,p3)] = do prob <- getStdRandom (randomR (0.0,1.0))
--                                    if (prob <= p3) 
--                                            then do return(c3)
--                                            else if ((prob > p3) && (prob <= (p3+p2))) 
--                                                    then do return(c2)
--                                                    else if ((prob > (p3+p2)) && (prob <= 1))
--                                                            then do return(c1)
--                                                            else return(c1)
--pick [(c1,p1),(c2,p2)]         = do prob <- getStdRandom (randomR (0.0,1.0))
--                                    if (prob >= p1) 
--                                            then do return(c1)
--                                            else return(c2)
--pick [(c1,p1)]                 = do return(c1)    

sortPropChordList :: PropChordList -> PropChordList
sortPropChordList l = sortBy sortGT l    

sortGT :: Ord a => (t, a) -> (t1, a) -> Ordering
sortGT (_, b1) (_, b2)
  | b1 < b2 = GT
  | b1 >= b2 = LT

transf :: PropChordList -> PropChordList
transf []         = []
transf [(c1,p1),(c2,p2),(c3,p3)] = [(c1,np1),(c2,np2),(c3,np3)] where 
                                           np1 = (p1/(p1+p2+p3))
                                           np2 = (p2/(p1+p2+p3))
                                           np3 = (p3/(p1+p2+p3))                                                                                      
transf [(c1,p1),(c2,p2)] = [(c1,np1),(c2,np2)] where
                                           np1 = (p1/(p1+p2))
                                           np2 = (p2/(p1+p2))
transf [(c,_)] = [(c,1)]

propChords :: Key -> [(Bar,[(Int,Octave,Duration,ChordList)])] -> [(Bar,[(Int,Octave,Duration,PropChordList)])]
propChords k []         = []
propChords k ((b,m):ms) = (b,(propMel k m)):propChords k ms where
        propMel :: Key -> [(Int,Octave,Duration,ChordList)] -> [(Int,Octave,Duration,PropChordList)] 
        propMel _ []             = []
        propMel k1 ((i,o,d,l):ms) = (i,o,d,(prop k l i)):propMel k1 ms where
                prop :: Key -> ChordList -> Int -> PropChordList
                prop _ []     _ = []
                prop k2 (x:xs) i = (x,(getProp k x i)): prop k2 xs i

pretttyPrintMel :: (Show a) => (Show b) => [(Bar,[(a,Octave,Duration,b)])] -> IO ()
pretttyPrintMel ((b,l):xs) = do putStrLn ("    " ++ (show(b)) ++ " " ++ show (l) )
                                pretttyPrintMel xs
pretttyPrintMel []         = return ()     

pretttyPrintSeq :: (Show a) => [[a]] -> IO ()
pretttyPrintSeq (x:xs) = do putStrLn (show(x))
                            pretttyPrintSeq xs
pretttyPrintSeq []         = return ()     

barNotesToInts :: Key -> [(Bar,[MelodyTone])] -> [(Bar,[(Int,Octave,Duration,ChordList)])]
barNotesToInts k ((b,m):xs) = (b,(melodyToInt k m)): barNotesToInts k xs where
barNotesToInts _ []           = []      

melodyToInt :: Key -> [(Root,Octave,Duration,ChordList)] -> [(Int,Octave,Duration,ChordList)]
melodyToInt key ((n,o,d,l):notes) = ((diaDegToSemi $ (toScaleDegree key n)),o,d,l):melodyToInt key notes
melodyToInt key []              = [] 

findCandidateChords :: Key -> [(Bar,[(Int,Octave,Duration,ChordList)])] -> [(Bar,[(Int,Octave,Duration,ChordList)])]
findCandidateChords k []               = []  
findCandidateChords k ((b@(1,_),m):ms) = (b,(f k m)) : findCandidateChords k ms where
                                         f :: Key -> [(Int,Octave,Duration,ChordList)] -> [(Int,Octave,Duration,ChordList)]
                                         -- if the list contains V or I, then leave only those for the first note, otherwise leave list intact
                                         f k ((i,o,d,l):xs) = (i,o,d,(onlyVOrI' $ findChordsByInt2 i k 0)): f' k xs
                                         f _ []           = []
                                         -- rest of the melody tones for the first bar are harmonised without constraint
                                         f' :: Key -> [(Int,Octave,Duration,ChordList)] -> [(Int,Octave,Duration,ChordList)]
                                         f' k ((i,o,d,l):xs) = (i,o,d,(findChordsByInt2 i k 0)): f' k xs
                                         f' _ []           = []                                                                                   
findCandidateChords k ((b@(n,d),m):ms) | (n==d)    =  (b,(f2 k m)) : findCandidateChords k ms -- last bar
                                       | otherwise =  (b,(f' k m)) : findCandidateChords k ms where -- other bars
                                                        f' :: Key -> [(Int,Octave,Duration,ChordList)] -> [(Int,Octave,Duration,ChordList)]
                                                        f' k ((i,o,d,l):xs) = (i,o,d,(findChordsByInt2 i k 0)): f' k xs
                                                        f' _ []           = []   
                                                        f2 :: Key -> [(Int,Octave,Duration,ChordList)] -> [(Int,Octave,Duration,ChordList)]
                                                        f2 _ []                          = []
                                                        -- last note
                                                        f2 k ((i,o,d,l):[])              = [(i,o,d,(onlyVOrI' (findChordsByInt2 i k 0)))]         
                                                        -- n-1
                                                        f2 k ((i,o,d,l):xs@[(i2,o2,d2,l2)]) = (i,o,d,(checkIV (findChordsByInt2 i k 0) (findChordsByInt2 i2 k 0))): f2 k xs
                                                        -- intermediate
                                                        f2 k ((i,o,d,l):xs)              = (i,o,d,(findChordsByInt2 i k 0)): f2 k xs

chordElem :: DiatonicDegree -> ChordList -> Bool               
chordElem _ []                                 = False
chordElem d ((Chord (Note (_) sd) _ _ _ _):xs) | (sd == d) = True
                                               | otherwise = chordElem d xs                                               
                                               
checkIV :: ChordList -> ChordList -> ChordList
checkIV [] _                                = [] -- should never happen
checkIV _ []                                = [] -- idem
checkIV this@(x:xs) next@(y:ys)              | (chordElem V next) = onlyIV' this
                                             | otherwise          = this

onlyIV' :: ChordList -> ChordList
onlyIV' l | (onlyIV l == [])   = l
          | otherwise          = onlyIV l

onlyIV :: ChordList -> ChordList
onlyIV []                                   = []
onlyIV (c@(Chord (Note (_) sd) _ _ _ _):xs) | (sd == IV)  = c:onlyVOrI xs
                                            | otherwise   = onlyVOrI xs  

onlyVOrI' :: ChordList -> ChordList
onlyVOrI' l | (onlyVOrI l == []) = l
            | otherwise          = onlyVOrI l

onlyVOrI :: ChordList -> ChordList
onlyVOrI []                                   = []
onlyVOrI (c@(Chord (Note (_) sd) _ _ _ _):xs) | ((sd == I) || (sd == V))  = c:onlyVOrI xs
                                              | otherwise                 = onlyVOrI xs 
onlyI :: ChordList -> ChordList
onlyI []                                   = []
onlyI (c@(Chord (Note (_) sd) _ _ _ _):xs) | ((sd == I))               = c:onlyI xs
                                           | otherwise                 = onlyI xs   

-- give every chord the number it occurs in the list and remove doubles
occursCheck :: (Eq a) => [a] -> [(a, Double)]
occursCheck l = occursCheck' l l where 
        occursCheck' :: (Eq a) => [a] -> [a] -> [(a, Double)]
        occursCheck' (x:xs) l2 = (x,fromIntegral(length(intersect l2 [x]))):occursCheck' xs l2
        occursCheck' []     _  = []

filterSkipNotes :: TimeSig -> Int -> [(Bar,[(Int,Octave,Duration,ChordList)])] -> [(Bar,[(Int,Octave,Duration,ChordList)])] 
filterSkipNotes t 0 l              = l
filterSkipNotes t s (((n,d),m):ms) = ((n,d),(filterSkipBar t s 0 m)):filterSkipNotes t s ms
filterSkipNotes t s []             = [] 
 
--filterSkipBar :: TimeSig -> Int -> Rational -> [(Int,Octave,Duration,ChordList)] -> [(Int,Octave,Duration,ChordList)]
--filterSkipBar (b1,b2) s a ((i,o,d,l):ms) | (a == (toRational(s) / toRational(b2)) || (a == 0)) = (i,o,d,l ):filterSkipBar (b1,b2) s (a+d) ms
--                                         | otherwise                                           = (i,o,d,[]):filterSkipBar (b1,b2) s (a+d) ms
--filterSkipBar s _ _       []             = []

filterSkipBar :: TimeSig -> Int -> Rational -> [(Int,Octave,Duration,ChordList)] -> [(Int,Octave,Duration,ChordList)]
-- time sig, skip val, acc, mel
-- s = 1, a = 0.5
filterSkipBar _       _ _ []             = []
filterSkipBar (b1,b2) s 0 ((i,o,d,l):ms) = (i,o,d,l ):filterSkipBar (b1,b2) s d ms
filterSkipBar (b1,b2) s a ((i,o,d,l):ms) | (elem a $ dr2 s $ strongList (b1,b2))    = (i,o,d,l ):filterSkipBar (b1,b2) s d ms                                         
                                         | otherwise                                = (i,o,d,[]):filterSkipBar (b1,b2) s (a+d) ms
                                         -- | (a == 0)                                = (i,o,d,l ):filterSkipBar (b1,b2) s (a+d) ms
                                         -- | (a == (toRational(s%(floor b2))))       = (i,o,d,l ):filterSkipBar (b1,b2) s d ms                                         

keyToString :: Key -> String
keyToString  (Key r m) = show r ++ modeToString m

modeToString :: Mode -> String
modeToString m | (null $ show m) = ":maj"
               | otherwise       = ":min"

--------------------------------------------------------------------------------
-- Command-line arguments
--------------------------------------------------------------------------------

data MyArgs = SourceInputString | SourceInputFile | TargetInputFile 
            | InputDir | OpMode | Print | MaxErrorRate | BinaryOut | BinaryIn
            | PrintIns | Grammar | TargetKeyInputFile
            | SourceKeyInputFile | AnnotationKeyInputDir| GroundTruthInputFile 
            | GroundTruthInputDir
            deriving (Eq, Ord, Show)

myArgs :: [Arg MyArgs]
myArgs = [
          Arg { argIndex = MaxErrorRate,
                argAbbr  = Just 'e',
                argName  = Just "max-err",
                argData  = argDataOptional "float" ArgtypeFloat,
                argDesc  = "Ignore pieces with higher error rate for diff"
              },
          Arg { argIndex = SourceInputString,
                argAbbr  = Just 'c',
                argName  = Just "chords",
                argData  = argDataOptional "string" ArgtypeString,
                argDesc  = "Input Chord Sequence to parse"
              },
          Arg { argIndex = SourceInputFile,
                argAbbr  = Just '1',
                argName  = Just "sfile",
                argData  = argDataOptional "filepath" ArgtypeString,
                argDesc  = "Input file (source for diff)"
              },
          Arg { argIndex = TargetInputFile,
                argAbbr  = Just '2',
                argName  = Just "tfile",
                argData  = argDataOptional "filepath" ArgtypeString,
                argDesc  = "Input file (target for diff)"
              },
          Arg { argIndex = SourceKeyInputFile,
                argAbbr  = Just '3',
                argName  = Just "skey",
                argData  = argDataOptional "filepath" ArgtypeString,
                argDesc  = "Source input ground-truth key annotation file"
              },
          Arg { argIndex = TargetKeyInputFile,
                argAbbr  = Just '4',
                argName  = Just "tkey",
                argData  = argDataOptional "filepath" ArgtypeString,
                argDesc  = "Target input ground-truth key annotation file"
              },
          Arg { argIndex = GroundTruthInputFile,
                argAbbr  = Just '5',
                argName  = Just "gt",
                argData  = argDataOptional "filepath" ArgtypeString,
                argDesc  = "Source input ground-truth chord annotation file"
              },
          Arg { argIndex = GroundTruthInputDir,
                argAbbr  = Just 'a',
                argName  = Just "gt-dir",
                argData  = argDataOptional "dir" ArgtypeString,
                argDesc  = "Ground-truth Annotation directory"
              },
          Arg { argIndex = InputDir,
                argAbbr  = Just 'd',
                argName  = Just "dir",
                argData  = argDataOptional "dir" ArgtypeString,
                argDesc  = "Input directory (process all files within)"
              },
          Arg { argIndex = AnnotationKeyInputDir,
                argAbbr  = Just 'k',
                argName  = Just "key-dir",
                argData  = argDataOptional "dir" ArgtypeString,
                argDesc  = "Ground-truth Key annotation directory"
              },              
          Arg { argIndex = OpMode,
                argAbbr  = Just 'm',
                argName  = Just "mode",
                argData  = argDataRequired "string" ArgtypeString,
                argDesc  = 
                  "Mode: parse|stdiff|lces-s|lcessim|hanlign|\n" ++ 
  "                                 align|harm-cr|head-cr|simple-cr"
              },
          Arg { argIndex = Grammar,
                argAbbr  = Just 'g',
                argName  = Just "grammar",
                argData  = argDataRequired "string" ArgtypeString,
                argDesc  = "Grammar to use (jazz|pop)"
              },
          Arg { argIndex = Print,
                argAbbr  = Just 'p',
                argName  = Just "print",
                argData  = Nothing,
                argDesc  = "Set this flag to print a .png of the parse"
              },
          Arg { argIndex = PrintIns,
                argAbbr  = Just 's',
                argName  = Just "print-insertions",
                argData  = Nothing,
                argDesc  = "Set this flag to show inserted nodes"
              },
          Arg { argIndex = BinaryOut,
                argAbbr  = Just 'o',
                argName  = Just "out",
                argData  = argDataOptional "filepath" ArgtypeString,
                argDesc  = "Output binary file for parsing results"
              },
          Arg { argIndex = BinaryIn,
                argAbbr  = Just 'i',
                argName  = Just "in",
                argData  = argDataOptional "filepath" ArgtypeString,
                argDesc  = "Input binary file for matching"
              }
         ]

--------------------------------------------------------------------------------
-- HarmTrace Main
--------------------------------------------------------------------------------

data SourceData = Audio | ChordLab | Annotation
  
getDataType :: Args MyArgs -> SourceData
getDataType args
  | gotArg args SourceKeyInputFile     = Annotation -- for single files
  | gotArg args AnnotationKeyInputDir  = Annotation -- for directory reading
  | otherwise                          = ChordLab   -- add audio support

-- by default all post processing operations are executed             
defaultOpts :: [PPOption]
defaultOpts  = [ RemovePDPT    , RemoveInsertions
               , MergeDelChords, ExpandChordDurations ]    
  
err1, err2, err3, err4, err5 :: String
err1 = "Use a source file, or a directory."
err2 = "Use a source file and a target file, or a directory."
err3 = "Use a source file and optionally a target file."
err4 = "Use an audio-feature location and a ground-truth file, "++
       "or an audio-feature directory and a ground-truth directory."
err5 = "fharm: missing or invalid arguments\n" ++ 
       "fharm: fharm <number of harmonizations to generate> <a number n, every nth melody beat note gets a chord> <key of midi file in HarmTrace notation> <\"path of input midi file\"> \n" ++ 
       "fharm: if no time signature is found in the input midi file metadata, 4/4 is assumed."

inputKey :: [Char] -> Key
inputKey i 
  |(i == "Cb:Maj") = (Key (Note (Just Fl) C) MajMode)
  |(i == "Cb:Min") = (Key (Note (Just Fl) C) MinMode)
  |(i == "C:Maj")  = (Key (Note (Nothing) C) MajMode)
  |(i == "C:Min")  = (Key (Note (Nothing) C) MinMode)
  |(i == "C#:Maj") = (Key (Note (Just Sh) C) MajMode)
  |(i == "C#:Min") = (Key (Note (Just Sh) C) MinMode)
  |(i == "Db:Maj") = (Key (Note (Just Fl) D) MajMode)
  |(i == "Db:Min") = (Key (Note (Just Fl) D) MinMode)
  |(i == "D:Maj")  = (Key (Note (Nothing) D) MajMode)
  |(i == "D:Min")  = (Key (Note (Nothing) D) MinMode)
  |(i == "D#:Min") = (Key (Note (Just Sh) D) MinMode)
  |(i == "D#:Maj") = (Key (Note (Just Sh) D) MajMode)
  |(i == "Eb:Maj") = (Key (Note (Just Fl) E) MajMode)
  |(i == "Eb:Min") = (Key (Note (Just Fl) E) MinMode)
  |(i == "E:Maj")  = (Key (Note (Nothing) E) MajMode)
  |(i == "E:Min")  = (Key (Note (Nothing) E) MinMode)
  |(i == "E#:Maj") = (Key (Note (Just Sh) E) MajMode)
  |(i == "E#:Min") = (Key (Note (Just Sh) E) MinMode)
  |(i == "Fb:Maj") = (Key (Note (Just Fl) F) MajMode)
  |(i == "Fb:Min") = (Key (Note (Just Fl) F) MinMode)
  |(i == "F:Maj")  = (Key (Note (Nothing) F) MajMode)
  |(i == "F:Min")  = (Key (Note (Nothing) F) MinMode)
  |(i == "F#:Maj") = (Key (Note (Just Sh) F) MajMode)
  |(i == "F#:Min") = (Key (Note (Just Sh) F) MinMode)
  |(i == "Gb:Maj") = (Key (Note (Just Fl) G) MajMode)
  |(i == "Gb:Min") = (Key (Note (Just Fl) G) MinMode)
  |(i == "G:Maj")  = (Key (Note (Nothing) G) MajMode)
  |(i == "G:Min")  = (Key (Note (Nothing) G) MinMode)
  |(i == "G#:Maj") = (Key (Note (Just Sh) G) MajMode)
  |(i == "G#:Min") = (Key (Note (Just Sh) G) MinMode)  
  |(i == "A:Maj")  = (Key (Note (Just Fl) A) MajMode)
  |(i == "A:Min")  = (Key (Note (Just Fl) A) MinMode)
  |(i == "A:Maj")  = (Key (Note (Nothing) A) MajMode)
  |(i == "A:Min")  = (Key (Note (Nothing) A) MinMode)
  |(i == "A#:Maj") = (Key (Note (Just Sh) A) MajMode)
  |(i == "A#:Min") = (Key (Note (Just Sh) A) MinMode)
  |(i == "B:Maj")  = (Key (Note (Just Fl) B) MajMode)
  |(i == "B:Min")  = (Key (Note (Just Fl) B) MinMode)
  |(i == "B:Maj")  = (Key (Note (Nothing) B) MajMode)
  |(i == "B:Min")  = (Key (Note (Nothing) B) MinMode)
  |(i == "B#:Maj") = (Key (Note (Just Sh) B) MajMode)
  |(i == "B#:Min") = (Key (Note (Just Sh) B) MinMode)
  |otherwise       = (Key (Note (Nothing) N) MajMode)

--main :: IO ()
--main = do args <- parseArgsIO ArgsComplete myArgs
--          let mode = getRequiredArg args OpMode
--              grmS = getRequiredArg args Grammar
--              prnt = gotArg args Print
--              opts = if gotArg args PrintIns 
--                     then delete RemoveInsertions defaultOpts else defaultOpts
--              gram = case grmS of
--                       "jazz" -> GrammarEx Jazz
--                       "pop"  -> GrammarEx Pop
--                       s      -> usageError args ("Unknown grammar: " ++ s)
--          case mode of
--            "parse"     -> mainParse args opts prnt           gram
--            "stdiff"    -> mainMatch args opts False STDiff   gram
--            "hanlign"   -> mainMatch args opts prnt  HAnAlign gram
--            "lces-s"    -> mainMatch args opts prnt  LCESsize gram
--            "lcessim"   -> mainMatch args opts prnt  LCESsim  gram
--            "align"     -> mainMatch args opts prnt  Align    gram
--            "head-cr"   -> mainChordRec headAnnotator   args prnt
--            "simple-cr" -> mainChordRec simpleAnnotator args prnt
--            "harm-cr"   -> mainChordRec (harmonyAnnotator gram) args prnt
--            s           -> usageError args ("Unknown mode: " ++ s)


--mainParse :: Args MyArgs -> [PPOption] -> Bool -> GrammarEx -> IO ()
--mainParse args o p (GrammarEx g) =
--  do let cStr = getArgString args SourceInputString
--         mf1  = getArgString args SourceInputFile
--         ky   = getArgString args SourceKeyInputFile
--         bOut = getArgString args BinaryOut
--         mdir = getArgString args InputDir
--         kdir = getArgString args AnnotationKeyInputDir
--         sdat = getDataType args
--     case (sdat, cStr, mf1,mdir, p, ky, kdir) of
--       -- parse a string of chords
--       (ChordLab, Just c, Nothing, Nothing , False,Nothing,Nothing)  ->  
--       --                  gramm bool string  
--         do pr <- parseTreeVerb g o c 
--            mapM_ (print . gTreeHead) (parsedPiece pr)
--       -- and print a parsetree     
--       (ChordLab, Just c, Nothing, Nothing , True,Nothing,Nothing)   ->
--         do pr <- parseTree g o c 
--            let ts = map gTreeHead (parsedPiece pr)
--            _ <- printTreeHAn (pieceTreeHAn pr) (trimFilename ("pp" ++ c))
--            printTreeHAnF ts (trimFilename c) >> return ()
--       -- Parse one file, show full output
--       (ChordLab, Nothing, Just f1, Nothing , False,Nothing,Nothing) -> 
--         do pr  <- readFile f1 >>= parseTreeVerb g o
--            print (pieceTreeHAn pr)
--            mapM_ (print . gTreeHead) (parsedPiece pr)
--       (ChordLab, Nothing, Just f1, Nothing , True ,Nothing,Nothing) ->
--       --with post processing
--         do pr <- readFile f1 >>= parseTree g o
--            let ts = map gTreeHead (parsedPiece pr)
--            printTreeHAn (pieceTreeHAn pr) (f1 ++ ".postProc") >> return ()
--            printTreeHAnF ts f1 >> return () 
--       -- Parse all files in one dir, show condensed output 
--       (ChordLab, Nothing, Nothing, Just dir, False, Nothing, Nothing) ->
--         parseDir g o dir bOut
--       -- ** audio ground-truth annotation part **   
--       (Annotation, Nothing, Just f1,Nothing, False, Just kf, Nothing) -> 
--       -- parse a ground-truth annotation and its key and give verbose output      
--         do key <- readFile kf
--            pr  <- readFile f1 >>= parseAnnotationVerb g o key 
--            print (pieceTreeHAn pr)
--            mapM_ (print . gTreeHead) (take 10 $ parsedPiece pr) 
--       (Annotation, Nothing, Just f1, Nothing , True , Just kf,Nothing) ->
--       -- parse a ground-truth annotation and its key and print the parse 
--         do key <- readFile kf
--            pr  <- readFile f1 >>= parseAnnotation g o key 
--            let ts = map gTreeHead (parsedPiece pr)
--            printTreeHAn (pieceTreeHAn pr) (f1 ++ ".postProc") >> return ()
--            printTreeHAnF ts f1 >> return ()   
--       -- Parse all files in one dir, show condensed output 
--       (Annotation, Nothing, Nothing, Just dir,False,Nothing,Just kd) ->
--         parseAnnotationDir g o kd dir
--       -- Else throw error
--       _ -> usageError args err1
                   
--trimFilename :: String -> String
--trimFilename = filter (\x -> not (elem x ":*")) . concat . words . take 20 
                   
--mainMatch :: Args MyArgs -> [PPOption] -> Bool -> MatchMode -> GrammarEx -> IO ()
--mainMatch args o p m (GrammarEx g) =
--     do let cStr  = getArgString args SourceInputString
--            mf1   = getArg args SourceInputFile
--            mf2   = getArg args TargetInputFile
--            mdir  = getArg args InputDir
--            bIn   = getArgString args BinaryIn
--            me    = getArg args MaxErrorRate
--        case (cStr,mf1,mf2,mdir,p) of
--          -- Parse source and target file, show full output
--          (_,Just f1, Just f2, Nothing, prnt)   -> 
--            do c1 <- readFile' f1
--               c2 <- readFile' f2
--               matchFiles o m prnt c1 c2 f1 f2
--          (Just c, Just f1, Nothing, Nothing, True) ->
--            matchFiles o m True c f1 (trimFilename c) (trimFilename f1)
--          -- match all files in one dir, show condensed output
--          (_,Nothing, Nothing, Just dir, False) -> dirMatch g o bIn m me dir 
--          _                                     -> usageError args err2

--matchFiles :: [PPOption] -> MatchMode -> Bool -> String -> String 
--           -> FilePath -> FilePath -> IO ()
--matchFiles o m prnt f1 f2 n1 n2 = 
--  -- should move to HarmTrace.IO.Main
--  let (ParseResult key1 toks1 _ ts1 _nr1 te1 pe1 _) 
--         = postProc o $ string2Piece Jazz f1
--      (ParseResult key2 toks2 _ ts2 _nr2 te2 pe2 _) 
--         = postProc o $ string2Piece Jazz f2
--  in
--  do  if not $ null te1 then showErrors "tokenizer 1: " te1 else putStr ""
--      if not $ null te2 then showErrors "tokenizer 2: " te2 else putStr ""
--      if not $ null pe1 then showErrors "parser 1: " pe1 else putStr ""
--      if not $ null pe2 then showErrors "parser 2: " pe2 else putStr ""
--      case (m,prnt) of
--        (STDiff,_)      -> print (diffChordsLen toks1 toks2)
--        (Align  ,False) -> print (getAlignDist key1 key2 toks1 toks2)
--        (Align  ,True ) -> do let (mat,v,t) = alignChordLab key1 key2 toks1 toks2 
--                              pPrintV t; print mat ; print v
                              
--        (HAnAlign,True ) -> do let (mat,v,t) = alignHAnChord ts1 ts2
--                               pPrintV t; print mat ; print v
--        -- quick and dirty LCES plotting (should move to HarmTrace.IO)
--        (LCESsize,True)  -> 
--           do printTreeHAn ts1 (n1 ++ ".postProc") >> return ()
--              printTreeHAn ts2 (n2 ++ ".postProc") >> return ()
--              printTreeHAnF (fst $ getLCES ts1 ts2) 
--                     (  (take 10 $ takeFileName n1) ++ ".vs." 
--                     ++ (take 10 $ takeFileName n2) ++ ".lces") >> return ()
--        _                -> error "Unimplemented." 

-- given a audio feature description files and groud-truth annotation, evaluates
-- a model-based chord labelling
--mainChordRec :: (Maybe [TimedData Key] -> AudioFeat -> ChordAnnotation) 
--             -> Args MyArgs -> Bool -> IO ()
--mainChordRec ann args p = 
--     do let af     = getArgString args SourceInputFile
--            gt     = getArg args GroundTruthInputFile
--            key    = getArg args SourceKeyInputFile
--            afdir  = getArg args InputDir
--            gtdir  = getArg args GroundTruthInputDir
--            keydir = getArg args AnnotationKeyInputDir
--        case (gt,af,key,gtdir,afdir,keydir,p) of
--          -- evaluates a single audio feature set
--          (Just g, Just a, Nothing, Nothing, Nothing, Nothing, prnt) -> 
--            evaluateLabeling ann prnt g a Nothing >>= print 
--          -- evaluates a single audio feature set and key annotation
--          (Just g, Just a, k, Nothing, Nothing, Nothing, prnt) -> 
--            evaluateLabeling ann prnt g a k >>= print   
--          -- evaluates a directory with audio features
--          (Nothing, Nothing, Nothing, Just gd, Just ad, Nothing, _ ) -> 
--            batchLabeling ann gd ad Nothing
--          -- evaluates a directory with audio features and key annotations
--          (Nothing, Nothing, Nothing, Just gd, Just ad, k, _ ) -> 
--            batchLabeling ann gd ad k
--          _                                        -> usageError args err4   
